export interface IJsonPatch {
    op: "replace" | "add" | "remove";
    path: string;
    value?: any;
}
export interface IReversibleJsonPatch extends IJsonPatch {
    oldValue: any;
}
/**
 * escape slashes and backslashes
 * http://tools.ietf.org/html/rfc6901
 */
export declare function escapeJsonPath(str: string): string;
/**
 * unescape slashes and backslashes
 */
export declare function unescapeJsonPath(str: string): string;
/**
 * Generates a json-path compliant json path from path parts
 *
 * @export
 * @param {string[]} path
 * @returns {string}
 */
export declare function joinJsonPath(path: string[]): string;
/**
 * Splits and decodes a json path into several parts
 *
 * @export
 * @param {string} path
 * @returns {string[]}
 */
export declare function splitJsonPath(path: string): string[];
